/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { expect } from "chai";
import { PostgresAdapter } from "./postgres";
import { MonetAdapter, MonetConfig } from "./monet";
import { Adapter } from "../core/adapter";
import { Fixture as FixPostgres } from "../../test/postgres/fixture";
import { Fixture as FixMonet } from "../../test/monet/fixture";
import { ConfigParser } from "../util/configParser";
import { adapterScenario } from "../../test/scenario";

describe("Sql adapter", () => {

    // Initializing
    let config: any;
    let adapter: Adapter;
    let fixture;
    before(function (done): void {
        // Arrow function not used to get acces to this and skip the test
        const configPath =  process.env.BLENDB_SCHEMA_FILE;
        config = ConfigParser.parse(configPath);
        if (config.adapter === "postgres") {
            fixture = new FixPostgres(config.connection);
            fixture.load(config.loadViews, (err) => {
                if (err) {
                    throw err;
                }
                adapter = new PostgresAdapter(config.connection);
                done();
            });
        }
        else if (config.adapter === "monet") {
            fixture = new FixMonet(config.connection);
            fixture.load(config.loadViews, (err) => {
                if (err) {
                    throw err;
                }
                let parsedConfig: MonetConfig = {
                    user: config.connection.user,
                    dbname: config.connection.database,
                    password: config.connection.password,
                    host: config.connection.host,
                    port: config.connection.port
                };
                adapter = new MonetAdapter(parsedConfig);
                done();
            });
        }
        else {
            this.skip();
        }
    });
    // Tests
    it("should get data from single materialized view", (done) => {
        let view = adapterScenario.materializedView;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should sort data from join of two views", (done) => {
        let view = adapterScenario.sortView;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(4);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
       });
    });
    it("should get data from single view (with sub-dimension)", (done) => {
        let view = adapterScenario.subDimensionView;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from join of one view", (done) => {
        let view = adapterScenario.joinWithOneView;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from view with a single clause", (done) => {
    let view = adapterScenario.filterWithEqual;
    adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(1);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from reduce with the same attributes of view", (done) => {
        let view = adapterScenario.reduceAsView;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from join into reduce with a single clause", (done) => {
        let view = adapterScenario.filterAverageBought;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(3);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from view with multiple clauses", (done) => {
        let view = adapterScenario.multipleClause;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(1);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from view with single clause, with more than one filter", (done) => {
        let view = adapterScenario.singleClause;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(4);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from view with no metrics", (done) => {
        let view = adapterScenario.joinWithNoMetrics;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from join with one metric and one view", (done) => {
        let view = adapterScenario.growOneView;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
    it("should get data from join with unrelated dimension", (done) => {
        let view = adapterScenario.JoinWithAncestors;
        adapter.getDataFromView(view, (err, result) => {
            expect(err).to.be.a("null");
            expect(result).to.be.an("array");
            expect(result).to.have.length(5);
            expect(result[0]).to.be.an("object");
            let keys: string[] = [];
            keys = keys.concat(view.metrics.map((item) => item.name));
            keys = keys.concat(view.dimensions.map((item) => item.name));
            result.forEach((row) => {
                expect(row).to.be.an("object");
                expect(row).to.have.all.keys(keys);
            });
            done();
        });
    });
});
