/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { Filter } from "./filter";
import { Hash } from "../util/hash";
import { Dimension } from "./dimension";
import { Metric } from "./metric";

/** Parameters used to create a Clause object. */
export interface ClauseOptions {
    /** Set of filters that form the clause. */
    filters: Filter[];
}

/** Set of restrictions applied to a query united by the operator OR. */
export class Clause {
    /** Hash of components that unique identify the clause. */
    public readonly id: string;
    /** Set of filters the form the clause. */
    public readonly filters: Filter[];
    /** Set of attributes affected by this clause. */
    public readonly targets: (Metric|Dimension)[];

    /**
     * Create a clause object.
     * @param options - Parameters required to create a clause object.
     */
    constructor (options: ClauseOptions) {
        this.filters = options.filters.map((i) => i);
        const t = this.filters.map((i) => i.target).sort((a, b) => {
            return (a.name < b.name) ? -1 : 1;
        });
        if (t.length > 0) {
            this.targets = [t[0]];
            for (let i = 1; i < t.length; ++i) {
                if (t[i - 1] !== t[i])  {
                    this.targets.push(t[i]);
                }
            }
        }
        else {
            this.targets = [];
        }
        const filtersIds = this.filters.map((item) => item.id);
        this.id = Hash.sha1(filtersIds.sort());
    }

    /**
     * Checks if the targets of this clause is a sub set of
     * the given coverage set.
     * In other words, this clause could be applied to a view
     * which has the coverage set as set of metrics and dimensions
     * @param coverage - Set of attributes which will be verified if
     * contain all targets in the clause.
     */
    public isCovered(coverage: (Metric|Dimension)[]): boolean {
        return this.targets.every((i) => coverage.some((j) => i.name === j.name));
    }
}
