/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { Dimension } from "./dimension";
import { Metric } from "./metric";
import { Hash } from "../util/hash";
import { DataType } from "../common/types";

/** Parameters used to create a filter object. */
export interface FilterOptions {
    /** Metric/Dimension that will be filtered. */
    target: Metric|Dimension;
    /** Operation applied to filter. */
    operator: FilterOperator;
    /** Constant value to be compared. */
    value: string;
}

/**
 * Parameters used to define filter object in the configuration file.
 * Also the string description of a filter.
 */
export interface StrFilterOptions {
    /** Metric/Dimension that will be filtered. */
    target: string;
    /** Operation applied to filter. */
    operator: string;
    /** Constant value to be compared. */
    value: string;
}

/** Available operations to be used in filters */
export enum FilterOperator {
    /** Used as error code. So suitable operation found. */
    NONE,
    /** Equality (==). */
    EQUAL,
    /** Not equality (!=). */
    NOTEQUAL,
    /** Greater than (>). */
    GREATER,
    /** Lower than (<). */
    LOWER,
    /** Greater or equal than( >=). */
    GREATEREQ,
    /** Lower or equal than (<=). */
    LOWEREQ
}

export class Filter {
    /** Hash of components that unique identify the filter. */
    public readonly id: string;
    /** Metric/Dimension that will be filtered. */
    public readonly target: Metric|Dimension;
    /** Constant value to be compared. */
    public readonly operator: FilterOperator;
    /** Constant value to be compared. */
    public readonly value: string;
    /** Constant that informs if the target type allow the operator. */
    public readonly isValid: boolean;

    /**
     * Create a filter.
     * @param options - Parameters required to create a filter.
     */
    constructor (options: FilterOptions) {
        this.target = options.target;
        this.operator = options.operator;
        this.value = options.value;
        this.id = Hash.sha1(options.target.name + options.operator + options.value);
        this.isValid = Filter.isTypeValid(options);
    }

    /**
     * Parse a string to enum(Filter Operator).
     * @param op - Filter operator in string format.
     */
    public static parseOperator(op: string): FilterOperator {
        switch (op) {
            case "==":
                return FilterOperator.EQUAL;
            case "!=":
                return FilterOperator.NOTEQUAL;
            case ">":
                return FilterOperator.GREATER;
            case "<":
                return FilterOperator.LOWER;
            case ">=":
                return FilterOperator.GREATEREQ;
            case "<=":
                return FilterOperator.LOWEREQ;
            default:
                return FilterOperator.NONE;
        }
    }

    /**
     * Splits a continuous string in the string representation of the filter.
     * @param strFilter - Filter in string format.
     */
    public static segment(strFilter: string): StrFilterOptions {
        for (let i = 0; i < strFilter.length; ++i) {
            switch (strFilter[i]){
                case "=":
                    if (strFilter[i + 1] === "=") {
                        return {
                            target: strFilter.slice(0, i),
                            operator: "==",
                            value: strFilter.slice(i + 2)
                        };
                    }
                    break;
                case "!":
                    if (strFilter[i + 1] === "=") {
                        return {
                            target: strFilter.slice(0, i),
                            operator: "!=",
                            value: strFilter.slice(i + 2)
                        };
                    }
                    break;
                case ">":
                    if (strFilter[i + 1] === "=") {
                        return {
                            target: strFilter.slice(0, i),
                            operator: ">=",
                            value: strFilter.slice(i + 2)
                        };
                    }
                    else {
                        return {
                            target: strFilter.slice(0, i),
                            operator: ">",
                            value: strFilter.slice(i + 1)
                        };
                    }
                case "<":
                    if (strFilter[i + 1] === "=") {
                        return {
                            target: strFilter.slice(0, i),
                            operator: "<=",
                            value: strFilter.slice(i + 2)
                        };
                    }
                    else {
                        return {
                            target: strFilter.slice(0, i),
                            operator: "<",
                            value: strFilter.slice(i + 1)
                        };
                    }
                default:
                    break;
            }
        }

        return null;
    }

    /**
     * Verify if the target allows the operation.
     * @param op - Operator to be validated.
     */
    private static isTypeValid(op: FilterOptions): boolean {
        if (op.operator === FilterOperator.NONE) {
            return false;
        }

        if (op.operator === FilterOperator.GREATER ||
            op.operator === FilterOperator.LOWER ||
            op.operator === FilterOperator.GREATEREQ ||
            op.operator === FilterOperator.LOWEREQ) {
            if (op.target.dataType === DataType.DATE || op.target.dataType === DataType.INTEGER) {
                return true;
            }

            else {
                return false;
            }
        }

        else {
            return true;
        }
    }

}
