/*
 * Copyright (C) 2016 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blend.
 *
 * blend is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blend is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blend.  If not, see <http://www.gnu.org/licenses/>.
 */

import { Dimension } from "./dimension";
import { Metric } from "./metric";
import { Hash } from "../util/hash";
import { Clause } from "./clause";
import { Operation, Opcode } from "../common/expression";

/** Information required to load fixture data into the view. */
export interface LoadView {
    /** View where the fixture data will be load. */
    view: View;
    /** Path to a file with fixture data. */
    data: string;
}

/** Parameters used to create a view object. */
export interface ViewOptions {
    /** Set of metrics contained into the view. */
    metrics: Metric[];
    /** Set of dimension contained into the view. */
    dimensions: Dimension[];
    /** Inform if this view cannot be generated by other views. */
    origin: boolean;
    /** Set of clauses applied to the view. */
    clauses?: Clause[];
    /** List with the sorting order of the attributes. */
    sort?: (Metric|Dimension)[];
    /** Definition of how to create this view using other views. */
    operation?: Operation;
    /** Defines if views's name will be its alias or automatically generated. */
    aliasAsName?: boolean;
    /** Fake identifier (human understandable) of the view. */
    alias?: string;
}

/**
 * Representation of data location in the BlenDB perspective.
 * Readable data are represented by metrics and dimensions and
 * these objects can only be found in views.
 */
export class View {
    /** Hash of components that unique identify the view. */
    public readonly id: string;
    /** Set of metrics contained into the view. */
    public readonly metrics: Metric[];
    /** Set of dimension contained into the view. */
    public readonly dimensions: Dimension[];
    /** Set of clauses applied to the view. */
    public readonly clauses: Clause[];
    /** List with the sorting order of the attributes. */
    public readonly sort: (Metric|Dimension)[];
    /** Inform if this view cannot be generated by other views. */
    public readonly origin: boolean;
    /** Definition of how to create this view using other views. */
    public readonly operation: Operation;
    /** View identifier, used on query generation as "table" name */
    public readonly name: string;

    /**
     * Create a view.
     * @param options - Parameters required to create a view.
     */
    constructor (options: ViewOptions) {
        this.metrics = options.metrics.sort();
        this.dimensions = options.dimensions.sort();
        this.clauses = (options.clauses) ? options.clauses.sort() : [];
        this.sort = (options.sort) ? options.sort : [];
        this.origin = options.origin || false;
        if (options.operation) {
            this.operation = options.operation;
        }

        else {
            this.operation = {
                opcode: Opcode.PUSH,
                values: []
            };
        }

        // calculate the id of the view based on it's metrics and dimensions
        const metNames = this.metrics.map((metric) => metric.name);
        const dimNames = this.dimensions.map((dimension) => dimension.name);
        const clausesIds = this.clauses.map((clause) => clause.id);
        const join = metNames.concat(dimNames).concat(clausesIds).sort();
        this.id = Hash.sha1(join);

        if ( options.aliasAsName ) {
            this.name = "\"" + options.alias + "\"";
        } else {
            this.name =  "view_" + this.id;
        }

    }

}
