/*
 * Copyright (C) 2018 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * A representation of a dependency relation.
 * Equivalent to a edge in a dependency graph.
 */
export interface TsortDep {
    /** The value (vertex). */
    value: string;
    /** The value which this depends of. */
    dependOf?: string;
}

/**
 * A dependency graph.
 */
interface Graph {
    /**
     * Dictionary indexed by vertices which
     * returns a list of the adjacent vertex.
     */
    [key: string]: string[];
}

interface VertexMark {
    /**
     * Dictionary indexed by vertices which
     * checks if the vertex is already sorted.
     */
    [key: string]: boolean;
}

/**
 * Dependency sorting methods. Creates a dependency
 * graph and use the topological sorting algorithm
 * putting all the dependencies before the dependent.
 */
export class Tsort {
    /**
     * Topologicaly sort strings in the TSortDep format.
     * @param deps - Dependency relations (edges in the
     * dependency graph).
     */
    public static dependencies(deps: TsortDep[]): string[]{
        let graph: Graph = {};
        let mark: VertexMark = {};
        for (let i = 0; i < deps.length; ++i) {
            let vertex = deps[i].dependOf;
            if (!graph[deps[i].value]) {
                graph[deps[i].value] = [];
                mark[deps[i].value] = false;
            }

            if (vertex) {
                if (!graph[vertex]) {
                    graph[vertex] = [deps[i].value];
                    mark[vertex] = false;
                }

                else {
                    graph[vertex].push(deps[i].value);
                }
            }

        }

        let output: string[] = [];
        for (let vertex of Object.keys(graph)) {
            if (!mark[vertex]) {
                Tsort.markVertex(vertex, graph, mark, output);
            }
        }
        return output;
    }

    /**
     * Recursive topological sorting algorithm strarting from
     * a given vertex
     * @param vertex - Topological sorting root.
     * @param graph - Dependency graph.
     * @param mark - Mark indicatiing if the vertex is already sorted.
     * @param output - The topological sorting.
     */
    private static markVertex(vertex: string, graph: Graph,
                              mark: VertexMark, output: string[]): void {
        if (mark[vertex]) {
            return;
        }

        mark[vertex] = true;
        for (let i = 0; i < graph[vertex].length; ++i) {
            Tsort.markVertex(graph[vertex][i], graph, mark, output);
        }
        output.unshift(vertex);
        return;
    }
}
