/*
 * Copyright (C) 2017 Centro de Computacao Cientifica e Software Livre
 * Departamento de Informatica - Universidade Federal do Parana
 *
 * This file is part of blendb.
 *
 * blendb is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * blendb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with blendb.  If not, see <http://www.gnu.org/licenses/>.
 */

import { ConfigParser } from "../src/util/configParser";
import { Metric } from "../src/core/metric";
import { Dimension } from "../src/core/dimension";
import { View } from "../src/core/view";
import { Filter, FilterOperator } from "../src/core/filter";
import { Clause } from "../src/core/clause";
import { AggregationType, RelationType , DataType} from "../src/common/types";
import { ViewHandler } from "../src/util/viewHandler";
import { EngineScenario, AdapterScenario, DataCtrlScenario } from "../src/util/scenarioHandler";
import { Query, QueryOpts } from "../src/common/query";

const configPath =  process.env.BLENDB_SCHEMA_FILE;
const config = ConfigParser.parse(configPath);

const mets : {[key: string]: Metric} = {};
for (let i in config.metrics) {
    let met = config.metrics[i];
    mets[met.name] = met;
}

const dims : {[key:string]: Dimension} = {};
for (let i in config.dimensions) {
    let dim = config.dimensions[i];
    dims[dim.name] = dim;
}

const views : {[key:string]: View} = {};
for (let i in config.buildViews){
    let view = config.buildViews[i];
    views[view.alias] = view.view;
}

const wrongMet = new Metric({
    name: "met:this:is:just:a:test",
    aggregation: AggregationType.COUNT,
    dataType: DataType.INTEGER
});

const wrongDim = new Dimension({ name: "dim:this:is:just:a:test", dataType: DataType.INTEGER });

/**
 * Create new filters to use in clause and test
 * the clauses
 */
const filters: { [key: string]: Filter } = {
    "equal" : new Filter({
        target : dims["dim:product:validity"],
        operator : FilterOperator.EQUAL,
        value: "2018-05-10"
    }),
    "lower" : new Filter({
        target : dims["dim:product:validity"],
        operator : FilterOperator.LOWER,
        value: "2018-05-10"
    }),
    "greater" : new Filter({
        target : dims["dim:product:validity"],
        operator : FilterOperator.GREATER,
        value: "2018-05-10"
    }),
    "greaterq" : new Filter({
        target : dims["dim:product:validity"],
        operator : FilterOperator.GREATEREQ,
        value: "2018-06-10"
    }),
    "lowerq" : new Filter({
        target : dims["dim:product:validity"],
        operator : FilterOperator.LOWEREQ,
        value: "2018-06-10"
    }),
    "noteq" : new Filter({
        target: dims["dim:product:validity"],
        operator: FilterOperator.NOTEQUAL,
        value: "2018-08-23"
    }),
    "clientRegistered" : new Filter({
        target: dims["dim:sell:registered"],
        operator: FilterOperator.EQUAL,
        value: "false"
    }),
    "averageBought": new Filter({
        target: mets["met:sell:avg:quantity"],
        operator: FilterOperator.GREATEREQ,
        value: "10"
    }),
    "equalFilterView": new Filter({
        target: dims["dim:seller:status"],
        operator: FilterOperator.EQUAL,
        value: "active"
    })
}

const subdims : {[key:string]: Dimension} = {
    "subdims_day" : new Dimension({
        name: "subdims_day",
        dataType: DataType.INTEGER,
        parent: dims["dim:sell:datein"],
        relation: RelationType.DAY
    }),
    "subdims_month" : new Dimension({
        name: "subdims_month",
        dataType: DataType.INTEGER,
        parent: dims["dim:sell:datein"],
        relation: RelationType.MONTH
    }),
    "subdims_year" : new Dimension({
        name: "subdims_year",
        dataType: DataType.INTEGER,
        parent: dims["dim:sell:datein"],
        relation: RelationType.YEAR
    }),
    "subdims_none" : new Dimension({
        name: "subdims_none",
        dataType: DataType.INTEGER,
        parent: null,
        relation: RelationType.NONE
    })
}

const clauses: { [key: string]: Clause }  = {
    "lastDay":
    new Clause({filters: [filters["equal"]]}),
    "undefined":
    new Clause({filters: [filters["noteq"]]}),
    "unexpired":
    new Clause({filters: [filters["greater"]]}),
    "expired":
    new Clause({filters: [filters["lower"]]}),
    "unexpiredLastDay":
    new Clause({filters: [filters["greaterq"]]}),
    "expiredlastDay":
    new Clause({filters: [filters["lowerq"]]}),
    "clientRegistered":
    new Clause({filters: [filters["clientRegistered"]]}),
    "averageBought":
    new Clause({filters: [filters["averageBought"]]}),
    "expiredAndAverage":
    // or between filteres => (A) || (B)
    new Clause ({filters: [filters["lower"],filters["averageBought"]]}),
    "equalClauseView":
    new Clause ({filters: [filters["equalFilterView"]]})
}


let qOpts : {[key: string]: QueryOpts} = {
    "JoinWithAncestors":
    {
        metrics: [mets["met:sell:count:quantity"]],
        dimensions: [dims["dim:seller:id"],dims["dim:provider:id"]]
    },
    "joinWithNoMetrics":
    {
        metrics: [],
        dimensions: [dims["dim:product:name"],dims["dim:seller:name"]]
    },
    "growOneView":
    {
        metrics: [mets["met:seller:min:age"]],
        dimensions: [dims["dim:seller:name"],dims["dim:seller:sex"]]
    },
    "multipleClause":
    {
        metrics: [mets["met:sell:avg:quantity"]],
        dimensions: [dims["dim:sell:datein"],dims["dim:seller:name"],
        dims["dim:client:name"]],
        // and between filters => (A) and (B)
        clauses: [clauses["expired"],clauses["averageBought"]],
    },
    "singleClause":
    {
        metrics: [mets["met:sell:avg:quantity"]],
        dimensions: [dims["dim:sell:datein"],dims["dim:seller:name"],
        dims["dim:client:name"]],
        clauses: [clauses["expiredAndAverage"]],
    },
    "equalfilter":
    {
        metrics: [],
        dimensions: [dims["dim:client:name"],dims["dim:product:validity"]],
        clauses: [clauses["lastDay"]],
    },
    "withSortView:0":
    {
        metrics: [mets["met:sell:sum:quantity"]],
        dimensions: [dims["dim:client:name"]],
        sort: [mets["met:sell:sum:quantity"]]
    },
    "withSortView:1":
    {
        metrics: [mets["met:sell:sum:quantity"]],
        dimensions: [dims["dim:client:id"]]
    },
    "subDimView:0":
    {
        metrics: [],
        dimensions : [dims["dim:sell:datein"],subdims["subdims_day"],
        subdims["subdims_month"],subdims["subdims_year"]]
    },
    "subDimView:1":
    {
        metrics: [],
        dimensions: [dims["dim:sell:datein"]]
    },
    "joinOneView":
    {
        metrics: [mets["met:product:avg:pricein"]],
        dimensions: [],
    },
    "reduceAsView":
    {
        metrics: [mets["met:sell:sum:quantity"],mets["met:sell:avg:quantity"],
    mets["met:sell:count:quantity"]],
        dimensions: [dims["dim:sell:registered"], dims["dim:product:id"],
    dims["dim:seller:id"], dims["dim:client:id"],dims["dim:sell:datein"]],
        sort: [mets["met:sell:sum:quantity"]]
    },
    "clientAverageBought:0":
    {
        metrics: [mets["met:sell:avg:quantity"]],
        dimensions: [dims["dim:client:name"]],
        clauses: [clauses["averageBought"]]
    },
    "clientAverageBought:1":
    {
        metrics: [mets["met:sell:avg:quantity"]],
        dimensions: [dims["dim:client:name"], dims["dim:seller:id"]]
    },
    "queryNoParent":
    {
        metrics: [mets["met:sell:count:quantity"]],
        dimensions: [subdims["subdims_none"]]
    },
    "queryMetsDims":
    {
        metrics : Object.keys(mets).map((key) => mets[key]),
        dimensions : Object.keys(dims).map((key) => dims[key])
    },
    "queryNoMets": 
    {
        metrics: [wrongMet],
        dimensions: [dims["dim:product:name"]]
    },
    "queryNoDims": 
    {
        metrics: [mets["met:buyout:min:quantity"]],
        dimensions: [wrongDim]
    },
    "queryProduct":
    {
        metrics: [mets["met:product:avg:pricein"], mets["met:product:max:pricein"], mets["met:product:min:pricein"],
        mets["met:product:avg:priceout"],mets["met:product:max:priceout"],mets["met:product:min:priceout"]],
        dimensions: [dims["dim:product:name"], dims["dim:product:validity"],dims["dim:product:id"]]
    },
    "queryActive":
    {
        metrics: [mets["met:seller:max:age"]],
        dimensions: [dims["dim:seller:name"],dims["dim:seller:status"]],
        clauses: [clauses["equalClauseView"]]
    },
    "querySubDim":
    {
        metrics : [],
        dimensions : [subdims["subdims_day"],subdims["subdims_month"]]
    },
    "wrongMet": 
    {
        metrics: [wrongMet],
        dimensions: [dims["dim:product:id"]]
    },
    "wrongDim": 
    {
        metrics: [mets["met:sell:avg:quantity"]],
        dimensions: [wrongDim]
    },
    "correct": 
    {
        metrics: [mets["met:buyout:avg:quantity"]],
        dimensions: [dims["dim:provider:id"]]
    },
    "clausal":
    {
        metrics: [mets["met:product:avg:pricein"]],
        dimensions: [dims["dim:product:name"],
        dims["dim:product:id"]]
    }
}

const queries : {[key: string]: Query} = {};
for(let i in qOpts){
    queries[i] = new Query(qOpts[i]);
}

const JoinWithAncestors = ViewHandler.growView(queries["JoinWithAncestors"],
    [views["view:Sell"]]);

const joinWithNoMetrics = ViewHandler.growView(queries["joinWithNoMetrics"],
    [views["view:Product"],
    views["view:Sell"],
    views["view:Seller"]]);

const growOneView = ViewHandler.growView(queries["growOneView"],
    [views["view:Seller"]]);

const multipleClause = ViewHandler.growView(queries["multipleClause"],
    [views["view:Sell"],
    views["view:Seller"],
    views["view:Client"],
    views["view:Product"]]);

const singleClause = ViewHandler.growView(queries["singleClause"],
    [views["view:Sell"],
    views["view:Seller"],
    views["view:Client"],
    views["view:Product"]]);

const equalfilter = ViewHandler.queryJoin(queries["equalfilter"],
    [views["view:Sell"],
    views["view:Client"],
    views["view:Product"]]);

const withSortView = ViewHandler.queryJoin(queries["withSortView:0"],
    [ViewHandler.queryReduce(queries["withSortView:1"],
    views["view:Sell"]),
    views["view:Client"]]);

const subDimView = ViewHandler.queryJoin(queries["subDimView:0"],
    [ViewHandler.queryReduce(queries["subDimView:1"],
    views["view:Sell"]),
    views["view:Sell"]]);

const joinOneView = ViewHandler.queryJoin(queries["joinOneView"],
    [views["view:Product"]]);

const reduceAsView = ViewHandler.queryReduce(queries["reduceAsView"],
    views["view:Sell"]);

const clientAverageBought = ViewHandler.queryReduce(queries["clientAverageBought:0"],
    ViewHandler.queryJoin(queries["clientAverageBought:1"],
    [views["view:Sell"],views["view:Client"]]));

const viewProduct = views["view:Product"];

const viewActiveSeller = views["view:ActiveSeller"];

// Exports

export const engineScenario: EngineScenario = {
    metrics: mets,
    dimensions: dims,
    wrongMet: wrongMet,
    wrongDim: wrongDim,
    subDimensions: subdims,
    views: views,
    queryMetsDims: queries["queryMetsDims"],
    queryNoMets: queries["queryNoMets"],
    queryNoDims: queries["queryNoDims"],
    queryProduct: queries["queryProduct"],
    viewProduct: viewProduct,
    queryActive: queries["queryActive"],
    viewActiveSeller: viewActiveSeller,
    querySubDim: queries["querySubDim"],
    queryNoParent: queries["queryNoParent"]
};

export const adapterScenario: AdapterScenario = {
    materializedView: views["view:Product"],
    sortView: withSortView,
    joinWithOneView: joinOneView,
    singleClause: singleClause,
    multipleClause: multipleClause,
    filterAverageBought: clientAverageBought,
    reduceAsView: reduceAsView,
    filterWithEqual: equalfilter,
    subDimensionView: subDimView,
    joinWithNoMetrics: joinWithNoMetrics,
    growOneView: growOneView,
    JoinWithAncestors:JoinWithAncestors,
};

export const dataCtrlScenario: DataCtrlScenario = {
    wrongMet: queries["wrongMet"],
    wrongDim: queries["wrongDim"],
    correct: queries["correct"],
    clausal: queries["clausal"],
};
